import argparse
import os
import sys
from getpass import getpass

import urllib3
from requests import request
from requests.auth import HTTPBasicAuth
from requests.exceptions import RequestException
from terminaltables import SingleTable

urllib3.disable_warnings()

COMMON_OPTIONS = [
    (
        ('-a', '--basic_auth',),
        dict(
            default=os.getenv('MYPASSWORDS_BASIC_AUTH'),
        ),
    ),
    (
        ('-k', '--insecure',),
        dict(
            action='store_true',
            default=os.getenv('MYPASSWORDS_INSECURE') == 'true' or False,
        ),
    ),
    (
        ('-u', '--url',),
        dict(
            default=os.getenv('MYPASSWORDS_URL') or 'http://127.0.0.1:8000',
        ),
    ),
]


DEFAULT_SUBOPTIONS = [
    (('site',), {}),
]


ADD_UPDATE_SUBOPTIONS = DEFAULT_SUBOPTIONS + [
    (('-l', '--login',), {}),
    (('--new_site',), {}),
]


class AlreadyExist(Exception):
    pass


class AmbiguousSite(Exception):
    pass


class NotFound(Exception):
    pass


class Cli:

    _data = None
    _id = None
    _method = None
    _passwords = None
    _settings = None

    @property
    def data(self):
        if self._data is not None:
            return self._data

        settings = self.settings
        command = settings.subparsers

        if command not in ['add', 'update']:
            return

        sys.stdout.write('Please enter a password to set:\n')

        self._data = {
            'password': getpass(),
            'site': settings.new_site or settings.site,
        }

        if settings.login:
            self._data['login'] = settings.login

        return self._data

    @property
    def id_(self):
        if self._id is not None:
            return self._id

        settings = self.settings
        command = settings.subparsers

        if command == 'search':
            return

        data = self.perform_request(params={'q': settings.site}, lookup=True)

        if command in ['add'] and not data:
            return
        if command in ['add'] and len(data) != 0:
            raise AlreadyExist
        elif command in ['delete', 'update', 'show'] and len(data) != 1:
            raise NotFound if not data else AmbiguousSite

        self._id = data[0]['id']
        return self._id

    @property
    def method(self):
        if self._method is not None:
            return self._method

        command = self.settings.subparsers

        if command == 'add':
            self._method = 'POST'
        elif command == 'delete':
            self._method = 'DELETE'
        elif command == 'update':
            self._method = 'PUT'
        else:
            self._method = 'GET'

        return self._method

    @property
    def passwords(self):
        if self._passwords is not None:
            return self._passwords

        command = self.settings.subparsers
        params = {} if command != 'search' else {'q': self.settings.site}

        self._passwords = self.perform_request(params=params) or []

        if not isinstance(self._passwords, list):
            self._passwords = [self._passwords]

        if command != 'show':
            self._passwords = [
                {k: v for k, v in d.items() if k != 'password'}
                for d in self._passwords
            ]

        return self._passwords

    @property
    def settings(self):
        if not self._settings:

            parser = argparse.ArgumentParser()

            for a, kw in COMMON_OPTIONS:
                parser.add_argument(*a, **kw)

            subparsers = parser.add_subparsers(dest='subparsers')

            parser_add = subparsers.add_parser('add')
            for a, kw in ADD_UPDATE_SUBOPTIONS:
                parser_add.add_argument(*a, **kw)

            parser_delete = subparsers.add_parser('delete')
            for a, kw in DEFAULT_SUBOPTIONS:
                parser_delete.add_argument(*a, **kw)

            parser_search = subparsers.add_parser('search')
            for a, kw in DEFAULT_SUBOPTIONS:
                parser_search.add_argument(*a, **kw)

            parser_show = subparsers.add_parser('show')
            for a, kw in DEFAULT_SUBOPTIONS:
                parser_show.add_argument(*a, **kw)

            parser_update = subparsers.add_parser('update')
            for a, kw in ADD_UPDATE_SUBOPTIONS:
                parser_update.add_argument(*a, **kw)

            self._settings, _ = parser.parse_known_args(sys.argv[1:])

            if self._settings.basic_auth \
                    and len(self._settings.basic_auth.split(':')) == 1:
                self._settings.password += ':{}'.format(getpass())

            if not self._settings.subparsers:
                parser.print_help()
                sys.exit(0)

        return self._settings

    def perform_request(self, params=None, lookup=False):

        settings = self.settings

        method = 'GET' if lookup else self.method
        id_ = None if lookup else self.id_
        data = None if lookup else self.data

        auth = None if not settings.basic_auth \
            else HTTPBasicAuth(*settings.basic_auth.split(':'))

        url = '{}/{}'.format(settings.url, 'passwords')
        if id_:
            url += '/{}'.format(id_)

        response = request(
            method,
            url,
            auth=auth,
            json=data,
            params=params,
            verify=not settings.insecure
        )

        try:
            response.raise_for_status()
        except RequestException as err:
            resp = err.response
            if resp.status_code in ['401', '403']:
                sys.stderr.write('Unauthorized\n')
            elif resp.status_code < 500:
                sys.stderr.write('Invalid request\n')
            else:
                sys.stderr.write('Server error\n')
            sys.exit(1)

        return None if method == 'DELETE' else response.json()


def main():

    try:
        passwords = Cli().passwords
    except AlreadyExist:
        sys.stdout.write('Already exist\n')
        sys.exit(1)
    except AmbiguousSite:
        sys.stdout.write('Ambiguous site\n')
        sys.exit(1)
    except NotFound:
        sys.stdout.write('Not found\n')
        sys.exit(1)
    except KeyboardInterrupt:
        sys.stdout.write('\n')
        return

    if not passwords:
        return

    heads = [c for c in sorted(passwords[0].keys())]
    lines = [[p[c] for c in heads] for p in passwords]

    table = [heads]
    table += lines

    sys.stdout.write(SingleTable(table).table)
    sys.stdout.write('\n')


if __name__ == "__main__":
    main()
